//------------------------------------------------------------------------------------------------

//
// File: polygon.cpp
// Author: Roman Lysecky
// Date Created: 28 November 2011
// 
// Description: Declaration of singly-linked list of LineSegment's representing a polygon.
// 
// Copyright: Roman Lysecky, Electrical and Computer Engineering, University of Arizona, 2011.
// 
// 

//------------------------------------------------------------------------------------------------

#include <iostream>
#include "linesegment.h"
#include "polygon.h"

//------------------------------------------------------------------------------------------------

using namespace std;

//------------------------------------------------------------------------------------------------

// Polygon constructor: initialize polygon to empty list 
Polygon::Polygon()
{
    head = 0;
    size = 0;
}

//------------------------------------------------------------------------------------------------

// 
// Polygon destructor: destroys the list by freeing all memory allocated for each line 
// segment returning the polygon to an empty list
//
Polygon::~Polygon()
{
    /* while the list of line segments is not empty, remove the element at the head of the list */
    while( size > 0 ) {
        remove_segment_next();
    }
}

//------------------------------------------------------------------------------------------------

// 
// inserts new line segment specified by (x1,y1), (x2, y2) into list after the given list element 
// location. If the list element location is NULL, inserts line segment at head of list 
//
void 
Polygon::insert_segment_next(float x1, float y1, float x2, float y2, LineSegment *location)
{
    LineSegment* new_segment;
    
    new_segment = new LineSegment(x1, y1, x2, y2);
    
    /* insertion at head of list */
    if( location == 0 ) {
        new_segment->next = head;
        head = new_segment;
    }
    /* insertion somewhere other than head */
    else {
        new_segment->next = location->next;
        location->next = new_segment;
    }
    size++;
}

//------------------------------------------------------------------------------------------------

/* 
 * removes the line segment after the the given list element location. If the list element location
 * is NULL, removes the head of the list
 */
void 
Polygon::remove_segment_next(LineSegment *location)
{
    LineSegment *old_segment;
    
    if( size == 0 ) return;
    
    if( location == 0 ) {
        old_segment = head;
        head = head->next;
    }
    else {
        if( location->next == 0 ) return;
        old_segment = location->next;
        location->next = location->next->next;
        
    }
    
    // free memory utilized for old_segment. this will call the LineSegment destrcutor.
    delete old_segment;
    size--;
}

//------------------------------------------------------------------------------------------------

/* returns the number of line segments within the polygon */
int 
Polygon::num_segments()
{
    return size;
}

//------------------------------------------------------------------------------------------------

/* returns pointer to next line segment within the list */
LineSegment *
Polygon::next_segment(LineSegment *segment) 
{
    if( segment == 0 ) return 0;
    return segment->next;
}


//------------------------------------------------------------------------------------------------

/* 
 * searches the polygon to find a segment that start (or ends) at the end of the provided 
 * current segment,returning a pointer to the connected segment if it exists. Otherwise, returns NULL
 *
 * If the connected segment ends (x2, y2) at the end point of the current segment, the start and
 * end of that segment will be swapped. This ensures the algorithm within is_valid_polygon()
 * always follows from the start (x1, y1) to the end (x2, y2) of every connected line segment.
 */
LineSegment * 
Polygon::find_next_connected_segment(LineSegment* fromSegment) 
{
    LineSegment *toSegment;
    float tempVal;
    
    toSegment = head;
    
    /* while we have not reached the end of the list of line segments */
    while (toSegment != 0 ) {
        
        /* only conisder those segments that are not the current segment */
        if( toSegment != fromSegment ) {
            
            /* if end (x2, y2) of from segment is equal to start of the to segment */
            if(fromSegment->x2 == toSegment->x1 &&
               fromSegment->y2 == toSegment->y1 ) {
                
                /* return the found segment */
                return toSegment;
            }
            /* if end (x2, y2) of from segment is equal to end of the to segment */
            else if (fromSegment->x2 == toSegment->x2 &&
                     fromSegment->y2 == toSegment->y2 ) {
                
                /* sawp the start and end of the to line segment */
                tempVal = toSegment->x2;
                toSegment->x2 = toSegment->x1;
                toSegment->x1 = tempVal;
                tempVal = toSegment->y2;
                toSegment->y2 = toSegment->y1;
                toSegment->y1 = tempVal;
                
                /* return found segment */
                return toSegment;
            }
        }
        toSegment = next_segment(toSegment);
    }
    
    /* did not find a conneced segment */
    return 0;
}

//------------------------------------------------------------------------------------------------

/* 
 * checks to determine if the specified polygon is a valid closed form polygon in which all segments 
 * are connected to another line segment
 *
 * NOTE: There is still a potential error with this implementation.
 */
bool 
Polygon::is_valid()
{
    LineSegment* segment;
    int segments_followed;
    
    /* start with first line segment in polygon */
    segment = head;
    segments_followed = 0;
    
    /* 
     * for each line segment found, search for another line segment that starts or ends
     * at the end point (x2, y2) of the current segment 
     *
     * continue as long as we have found another connected segment, we have not returned the
     * head of the list, and we have not followed more than the number of segments that are in
     * the polygon (i.e. make sure we don't get stuck in a cycle)
     */
    do {
        segment = find_next_connected_segment(segment);
        
        /* if the segment is not found, the polygon is invalid */
        if( segment == 0 ) return false;
        
        /* if segment was found, increase the count of segments we have followed */
        segments_followed++;
        
    } while (segment != head && segments_followed != size);
    
    /* 
     * if we returned to the head in exactly the number of segments that exist in the 
     * polygon, we have a valid polygon shape 
     */
    if(segment == head && segments_followed == size) return true;
    
    return false;
}

//------------------------------------------------------------------------------------------------

/* prints list of line segments within polygon */
void 
Polygon::display() {
    
    LineSegment *segment;
        
    segment = head;
    while ( segment != 0 ) {
        if (segment != head) cout << ", ";
        cout << "(" << segment->x1 << "," << segment->y1 << ")->(" <<
            segment->x2 << "," << segment->y2 << ")";
        
        segment = next_segment(segment);
    }
}

//------------------------------------------------------------------------------------------------

ostream& 
operator<< (ostream &os, Polygon &polygon)
{
    LineSegment *segment;
    
    segment = polygon.head;
    while ( segment != 0 ) {
        if (segment != polygon.head) cout << ", ";
        cout << *segment;
        
        segment = polygon.next_segment(segment);
    }
    
    return os;
}

//------------------------------------------------------------------------------------------------

